#include "io.h" /* io.h is implemented in the section "Moving the cursor" */

/* The I/O ports */
/* All the I/O ports are calculated relative to the data port.
 * This is because all serial ports (COM1, COM2, COM3, COM4) have their ports
 * in the same order, but they start at different values.
 */
#define SERIAL_COM1_BASE         0x3F8  /* COM1 base port address */
#define SERIAL_DATA_PORT(base)   (base) /* Data port for serial communication */
#define SERIAL_FIFO_COMMAND_PORT(base)   (base + 2)  /* FIFO command port */
#define SERIAL_LINE_COMMAND_PORT(base)   (base + 3)  /* Line command port */
#define SERIAL_MODEM_COMMAND_PORT(base)  (base + 4)  /* Modem command port */
#define SERIAL_LINE_STATUS_PORT(base)    (base + 5)  /* Line status port */

/* The I/O port commands */
/* SERIAL_LINE_ENABLE_DLAB:
 * Tells the serial port to expect first the highest 8 bits on the data port,
 * then the lowest 8 bits will follow.
 */
#define SERIAL_LINE_ENABLE_DLAB 0x80  /* Command to enable DLAB (Divisor Latch Access Bit) */

/**
 * serial_configure_baud_rate:
 * Sets the speed of the data being sent. The default speed of a serial port is 115200 bits/s.
 * The argument is a divisor of that number, hence the resulting speed becomes (115200 / divisor) bits/s.
 *
 * @param com The COM port to configure (COM1, COM2, etc.)
 * @param divisor The divisor value for setting the baud rate
 */
void serial_configure_baud_rate(unsigned short com, unsigned short divisor)
{
    /* Enable DLAB (Divisor Latch Access Bit) */
    outb(SERIAL_LINE_COMMAND_PORT(com), SERIAL_LINE_ENABLE_DLAB);
    
    /* Send the high byte of the divisor to the data port */
    outb(SERIAL_DATA_PORT(com), (divisor >> 8) & 0x00FF);
    
    /* Send the low byte of the divisor to the data port */
    outb(SERIAL_DATA_PORT(com), divisor & 0x00FF);
}

/**
 * serial_configure_line:
 * Configures the line of the given serial port. The port is set to have a
 * data length of 8 bits, no parity bits, one stop bit, and break control
 * disabled.
 *
 * @param com The serial port to configure (e.g., COM1, COM2, etc.)
 */
void serial_configure_line(unsigned short com)
{
    /* Bit:
     * | 7 | 6 | 5 4 3 | 2 | 1 0 |
     * Content:   | d | b | prty | s | dl |
     * Value:     | 0 | 0 | 0 0 0 | 0 | 1 1 | = 0x03
     *
     * Where:
     * - d = Data bits (8 bits)
     * - b = Break control (disabled)
     * - prty = Parity bits (none)
     * - s = Stop bits (1 stop bit)
     * - dl = Data length (8 bits, value 0x03)
     */
    outb(SERIAL_LINE_COMMAND_PORT(com), 0x03);  /* Set the line configuration */
}
